(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda ()'orange)]
           [("grape") ,(lambda () 'grape)]]))


; success-k : (val stream -> ...)

(define (parse nonterm table stream success-k fail-k)
  (local [(define defn (assq nonterm table))]
    (if defn
        (parse-alts (rest defn) table stream success-k fail-k)
        (error 'parse "no definition for ~a" nonterm))))

(define (parse-alts alts table stream success-k fail-k)
  (cond
    [(empty? alts) (fail-k)]
    [else (local [(define alt (first alts))]
            ;; parse-or
            (parse-seq (first alt) table stream
                       (lambda (lst stream)
                         (success-k (apply (second alt) lst)
                                    stream))
                       (lambda ()
                         (parse-alts (rest alts) table stream
                                     success-k fail-k))))]))
          
(define (parse-seq seq table stream success-k fail-k)
  (cond
    [(empty? seq) (success-k empty stream)]
    [else (parse-atom (first seq) table stream
                      (lambda (fst-lst stream)
                        (parse-seq (rest seq) table stream
                                   (lambda (rest-lst stream)
                                     (success-k (append fst-lst rest-lst)
                                                stream))
                                   (lambda ()
                                     
                                     (fail-k))))
                      (lambda ()
                        (fail-k)))]))


(define (parse-atom atom table stream success-k fail-k)
  (cond
    [(string? atom) (if (and (cons? stream)
                             (equal? (first stream) atom))
                        (success-k empty
                                   (rest stream))
                        (fail-k))]
    [(symbol? atom) (parse atom table stream 
                           (lambda (val stream)
                             (success-k (list val) stream))
                           fail-k)]))


(test (parse 'main table '("apple") list (lambda () false))
      (list 'apple empty))

(test (parse 'main table '("apple" "banana") list (lambda () false))
      (list 'apple '("banana")))

(test (parse 'main table '("orange" "grape" "!") list (lambda () false))
      (list '((orange grape) !) empty))

(test (parse 'main table '("orange" "grape" "orange" "!") list (lambda () false))
      '((orange grape orange) !))



