(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]
          [(fruits ".") ,(lambda (x) x)]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]
    [fruits [() ,(lambda () '())]
            [(fruit fruits) ,cons]]
    [fruits2 [() ,(lambda () '())]
             [(fruits2 fruit) ,(lambda (l v)
                                 (append l (list v)))]]))

(define (parse what table stream 
               success-k fail-k)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream 
                    success-k fail-k)
        (error 'parse "no definition for ~a" what))))

(define (parse-alts alts table stream success-k fail-k)
  (if (null? alts)
      (fail-k)
      ;; This could be abstracted out as "or"
      (local [(define (make-fst-success-k success-k fail-k)
                (lambda (val stream next-fst)
                  (success-k val stream
                             (lambda (success-k fail-k)
                               (next-fst (make-fst-success-k success-k fail-k)
                                         (make-fst-fail-k success-k fail-k))))))
              (define (make-fst-fail-k success-k fail-k)
                (lambda ()
                  (parse-alts (rest alts) table stream
                              success-k fail-k)))]
        (parse-case (first alts) table stream
                    (make-fst-success-k success-k fail-k)
                    (make-fst-fail-k success-k fail-k)))))

(define (make-success-k val->val success-k fail-k)
  (local [(define (make-my-success-k success-k fail-k)
            (lambda (val stream next)
              (success-k (val->val val) stream
                         (lambda (success-k fail-k)
                           (next (make-my-success-k success-k fail-k)
                                 fail-k)))))]
    (make-my-success-k success-k fail-k)))

(define (parse-case case table stream success-k fail-k)
  (local [(define seq (first case))
          (define handler (second case))]
    (parse-sequence seq table stream
                    (make-success-k (lambda (vals)
                                      (apply handler vals))
                                    success-k
                                    fail-k)
                    fail-k)))

(define (parse-sequence seq table stream success-k fail-k)
  (cond
    [(empty? seq) (success-k empty stream
                             (lambda (success-k fail-k)
                               (fail-k)))]
    [else 
     ;; This could be abstracted out as "and"
     (parse-item (first seq) table stream
                 (local [(define (make-fst-success-k success-k fail-k)
                           (lambda (fst-vals stream fst-next)
                             (parse-sequence 
                              (rest seq) table stream
                              (make-snd-success-k fst-vals fst-next success-k fail-k)
                              (lambda ()
                                (fst-next (make-fst-success-k success-k fail-k)
                                          fail-k)))))
                         (define (make-snd-success-k fst-vals fst-next success-k fail-k)
                           (lambda (snd-vals stream snd-next)
                             (success-k 
                              (append fst-vals snd-vals)
                              stream
                              (lambda (success-k fail-k)
                                (snd-next 
                                 (make-snd-success-k fst-vals fst-next success-k fail-k)
                                 (lambda ()
                                   (fst-next (make-fst-success-k success-k fail-k)
                                             fail-k)))))))]
                   (make-fst-success-k success-k fail-k))
                 fail-k)]))

(define (parse-item item table stream success-k fail-k)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (success-k empty (rest stream) 
                    (lambda (success-k fail-k)
                      (fail-k)))
         (fail-k))]
    [else
     (parse item table stream
            (make-success-k list success-k fail-k)
            fail-k)]))

(parse 'fruit-salad table '("orange" "orange" "!")
       (lambda (val stream again) val)
       (lambda () 'nope!))

(parse 'main table '("orange" "grape" "!")
       list
       (lambda () 'nope!))

(parse 'main table '("orange" "grape" "orange" ".")
       list
       (lambda () 'nope!))
